/* -*- mode: c++; coding: utf-8; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4; show-trailing-whitespace: t -*- vim:fenc=utf-8:ft=cpp:et:sw=4:ts=4:sts=4 */

#include <feel/feelmodels/fluid/fluidmecbase.hpp>

#include <feel/feelvf/vf.hpp>

#include <feel/feelmodels/modelvf/fluidmecstresstensor.hpp>


namespace Feel
{
namespace FeelModels
{

FLUIDMECHANICSBASE_CLASS_TEMPLATE_DECLARATIONS
void
FLUIDMECHANICSBASE_CLASS_TEMPLATE_TYPE::updatePicard( DataUpdateLinear & data ) const
{
    this->updateLinearPDE( data );
}

FLUIDMECHANICSBASE_CLASS_TEMPLATE_DECLARATIONS
void
FLUIDMECHANICSBASE_CLASS_TEMPLATE_TYPE::updateLinearPDE( DataUpdateLinear & data ) const
{
    using namespace Feel::vf;

    const vector_ptrtype& vecCurrentPicardSolution = data.currentSolution();
    sparse_matrix_ptrtype& A = data.matrix();
    vector_ptrtype& F = data.rhs();
    bool _BuildCstPart = data.buildCstPart();
    sparse_matrix_ptrtype& A_extended = data.matrixExtended();
    bool _BuildExtendedPart = data.buildExtendedPart();
    bool _doBCStrongDirichlet = data.doBCStrongDirichlet();

    bool BuildNonCstPart = !_BuildCstPart;
    bool BuildCstPart = _BuildCstPart;

    bool build_ConvectiveTerm = BuildNonCstPart;
    bool build_Form2TransientTerm = BuildNonCstPart;
    bool build_Form1TransientTerm = BuildNonCstPart;
    //bool build_SourceTerm = BuildNonCstPart;
    bool build_BoundaryNeumannTerm = BuildNonCstPart;
    bool build_StressTensorNonNewtonian = BuildNonCstPart;
    if ( this->timeStepBase()->strategy()==TS_STRATEGY_DT_CONSTANT )
    {
        build_Form2TransientTerm=BuildCstPart;
    }
    if (this->useFSISemiImplicitScheme())
    {
        build_StressTensorNonNewtonian = BuildCstPart;
        if ( this->solverName() == "Oseen" )
            build_ConvectiveTerm=BuildCstPart;
        build_Form2TransientTerm=BuildCstPart;
        build_Form1TransientTerm=BuildCstPart;
        //build_SourceTerm=BuildCstPart;
        build_BoundaryNeumannTerm=BuildCstPart;
    }

    std::string sc=(_BuildCstPart)?" (build cst part)":" (build non cst part)";
    this->log("FluidMechanics","updateLinearPDE", "start"+sc );
    this->timerTool("Solve").start();

    auto mesh = this->mesh();
    auto Xh = this->functionSpace();

    element_fluid_ptrtype fielCurrentPicardSolution;
    if ( this->solverName() == "Picard" )
    {
        fielCurrentPicardSolution = Xh->elementPtr();
        for ( size_type k=0;k<Xh->nLocalDofWithGhost();++k )
            fielCurrentPicardSolution->set( k,vecCurrentPicardSolution->operator()(/*rowStartInVector+*/k) );
    }

    auto const& U = this->fieldVelocityPressure();
    auto u = U.template element<0>();
    auto v = U.template element<0>();
    auto p = U.template element<1>();
    auto q = U.template element<1>();

    // strain tensor (trial)
    auto deft = sym(gradt(u));
    // density
    auto const& rho = this->densityViscosityModel()->fieldRho();
    // identity matrix
    auto const Id = eye<nDim,nDim>();

    //--------------------------------------------------------------------------------------------------//

    auto rowStartInMatrix = this->rowStartInMatrix();
    auto colStartInMatrix = this->colStartInMatrix();
    auto rowStartInVector = this->rowStartInVector();
    auto bilinearForm_PatternDefault = form2( _test=Xh,_trial=Xh,_matrix=A,
                                              _pattern=size_type(Pattern::DEFAULT),
                                              _rowstart=rowStartInMatrix,
                                              _colstart=colStartInMatrix );
    auto bilinearForm_PatternCoupled = form2( _test=Xh,_trial=Xh,_matrix=A,
                                              _pattern=size_type(Pattern::COUPLED),
                                              _rowstart=rowStartInMatrix,
                                              _colstart=colStartInMatrix );

    auto myLinearForm =form1( _test=Xh, _vector=F,
                              _rowstart=rowStartInVector );

    //--------------------------------------------------------------------------------------------------//
    this->timerTool("Solve").start();

    // stress tensor sigma : grad(v)
    if ( this->densityViscosityModel()->dynamicViscosityLaw() == "newtonian")
    {
        if ( BuildCstPart )
        {
            auto Sigmat = -idt(p)*Id + 2*idv(this->densityViscosityModel()->fieldMu())*deft;
            bilinearForm_PatternCoupled +=
                integrate( _range=elements(mesh),
                           _expr= inner(Sigmat,grad(v)),
                           _geomap=this->geomap() );
        }
    }
    else
    {
        if ( build_StressTensorNonNewtonian )
        {
            auto BetaU = ( this->solverName() == "Oseen" )? M_bdf_fluid->poly() : *fielCurrentPicardSolution;
            auto betaU = BetaU.template element<0>();
            auto myViscosity = Feel::vf::FeelModels::fluidMecViscosity<2*nOrderVelocity>(betaU,p,*this->densityViscosityModel());
            bilinearForm_PatternCoupled +=
                integrate( _range=elements(mesh),
                           _expr= 2*myViscosity*inner(deft,grad(v)),
                           _geomap=this->geomap() );
        }
        if ( BuildCstPart )
        {
            bilinearForm_PatternCoupled +=
                integrate( _range=elements(mesh),
                           _expr= -div(v)*idt(p),
                           _geomap=this->geomap() );
        }
    }

    // incompressibility term
    if ( BuildCstPart )
    {
        bilinearForm_PatternCoupled +=
            integrate( _range=elements(mesh),
                       _expr= -divt(u)*id(q),
                       _geomap=this->geomap() );
    }

    double timeElapsedStressTensor = this->timerTool("Solve").stop();
    this->log("FluidMechanics","updateLinearPDE","assembly stress tensor + incompressibility in "+(boost::format("%1% s") %timeElapsedStressTensor).str() );

    //--------------------------------------------------------------------------------------------------//
    // define pressure cst
    if ( this->definePressureCst() )
    {
        if ( this->definePressureCstMethod() == "penalisation" && BuildCstPart  )
        {
            double beta = this->definePressureCstPenalisationBeta();
            bilinearForm_PatternCoupled +=
                integrate( _range=elements(Xh->mesh()),
                           _expr=beta*idt(p)*id(q),
                           _geomap=this->geomap() );
        }
        if ( this->definePressureCstMethod() == "lagrange-multiplier" )
        {
            CHECK( this->startDofIndexFieldsInMatrix().find("define-pressure-cst-lm") != this->startDofIndexFieldsInMatrix().end() )
                << " start dof index for define-pressure-cst-lm is not present\n";
            size_type startDofIndexDefinePressureCstLM = this->startDofIndexFieldsInMatrix().find("define-pressure-cst-lm")->second;

#if defined(FLUIDMECHANICS_USE_LAGRANGEMULTIPLIER_ONLY_ON_BOUNDARY)
            auto therange=boundaryfaces(mesh);
#else
            auto therange=elements(mesh);
#endif
            auto lambda = M_XhMeanPressureLM->element();
            if (BuildCstPart)
            {
                form2( _test=Xh, _trial=M_XhMeanPressureLM, _matrix=A,
                       _rowstart=this->rowStartInMatrix(),
                       _colstart=this->colStartInMatrix()+startDofIndexDefinePressureCstLM ) +=
                    integrate( _range=therange,//elements(mesh),
                               _expr= id(p)*idt(lambda) /*+ idt(p)*id(lambda)*/,
                               _geomap=this->geomap() );

                form2( _test=M_XhMeanPressureLM, _trial=Xh, _matrix=A,
                       _rowstart=this->rowStartInMatrix()+startDofIndexDefinePressureCstLM,
                       _colstart=this->colStartInMatrix() ) +=
                    integrate( _range=therange,//elements(mesh),
                               _expr= + idt(p)*id(lambda),
                               _geomap=this->geomap() );
            }

#if defined(FLUIDMECHANICS_USE_LAGRANGEMULTIPLIER_MEANPRESSURE)
            if (BuildNonCstPart)
            {
                this->log("FluidMechanics","updateLinearPDE", "also add nonzero MEANPRESSURE" );
                form1( _test=M_XhMeanPressureLM, _vector=F,
                       _rowstart=this->rowStartInMatrix()+startDofIndexDefinePressureCstLM ) +=
                    integrate( _range=therange,//elements(mesh),
                               _expr= FLUIDMECHANICS_USE_LAGRANGEMULTIPLIER_MEANPRESSURE(this->shared_from_this())*id(lambda),
                               _geomap=this->geomap() );
            }
#endif
        } // if ( this->definePressureCstMethod() == "lagrange-multiplier" )
    } // if ( this->definePressureCst() )

    //--------------------------------------------------------------------------------------------------//
    // convection
    if ( this->modelName() == "Navier-Stokes" && build_ConvectiveTerm )
    {
        this->timerTool("Solve").start();

        CHECK( this->solverName() == "Oseen" || this->solverName() == "Picard" ) << "invalid solver name " << this->solverName();
        auto BetaU = ( this->solverName() == "Oseen" )? M_bdf_fluid->poly() : *fielCurrentPicardSolution;
        auto betaU = BetaU.template element<0>();
        if ( this->isMoveDomain() )
        {
#if defined( FEELPP_MODELS_HAS_MESHALE )
            bilinearForm_PatternDefault +=
                integrate( _range=elements(mesh),
                           _expr= idv(rho)*trans( gradt(u)*( idv(betaU) -idv( this->meshVelocity() )))*id(v),
                           _geomap=this->geomap() );
#endif
        }
        else
        {
            bilinearForm_PatternDefault +=
                integrate( _range=elements(mesh),
                           _expr= idv(rho)*trans( gradt(u)*idv(betaU) )*id(v),
                           _geomap=this->geomap() );
        }

        if ( this->doStabConvectionEnergy() )
        {
            bilinearForm_PatternCoupled +=
                integrate( _range=elements(mesh),
                           _expr= 0.5*idv(rho)*divt(u)*trans(idv(betaU))*id(v),
                           _geomap=this->geomap() );
        }

        double timeElapsedConvection = this->timerTool("Solve").stop();
        this->log("FluidMechanics","updateLinearPDE","assembly convection in "+(boost::format("%1% s") %timeElapsedConvection).str() );
    }
    else if ( this->modelName() == "Stokes" && build_ConvectiveTerm && this->isMoveDomain() )
    {
#if defined( FEELPP_MODELS_HAS_MESHALE )
        bilinearForm_PatternDefault +=
            integrate( _range=elements(mesh),
                       _expr= -idv(rho)*trans( gradt(u)*(idv( this->meshVelocity() )))*id(v),
                       _geomap=this->geomap() );
#endif
    }



    //--------------------------------------------------------------------------------------------------//
    //transients terms
    if (!this->isStationary())
    {
        if (build_Form2TransientTerm)
        {
            bilinearForm_PatternDefault +=
                integrate( _range=elements(mesh),
                           _expr= idv(rho)*trans(idt(u))*id(v)*M_bdf_fluid->polyDerivCoefficient(0),
                           _geomap=this->geomap() );
        }

        if (build_Form1TransientTerm)
        {
            auto Buzz = M_bdf_fluid->polyDeriv();
            auto buzz = Buzz.template element<0>();
            myLinearForm +=
                integrate( _range=elements(mesh),
                           _expr= idv(rho)*trans(idv(buzz))*id(v),
                           _geomap=this->geomap() );
        }
    }

    //--------------------------------------------------------------------------------------------------//
    // volume force
    this->updateSourceTermLinearPDE(F,BuildCstPart);
    // source given by user
    if ( M_haveSourceAdded && BuildNonCstPart)
    {
        myLinearForm +=
            integrate( _range=elements(mesh),
                       _expr= trans(idv(*M_SourceAdded))*id(v),
                       _geomap=this->geomap() );
    }
    if ( M_useGravityForce && BuildCstPart )
    {
        myLinearForm +=
            integrate( _range=elements(mesh),
                       _expr= idv(rho)*inner(M_gravityForce,id(v)),
                       _geomap=this->geomap() );
    }
    //--------------------------------------------------------------------------------------------------//
    // div u != 0
    if (!this->velocityDivIsEqualToZero() && BuildNonCstPart)
    {
        myLinearForm +=
            integrate( _range=elements(mesh),
                       _expr= -idv(this->velocityDiv())*id(q),
                       _geomap=this->geomap() );

        auto coeffDiv = (2./3.)*idv(this->densityViscosityModel()->fieldMu()); //(eps-2mu/3)
        myLinearForm +=
            integrate( _range=elements(mesh),
                       _expr= val(coeffDiv*gradv(this->velocityDiv()))*id(v),
                       _geomap=this->geomap() );

        if ( this->doStabConvectionEnergy() )
        {
            auto BetaU = M_bdf_fluid->poly();
            auto betaU = BetaU.template element<0>();
            myLinearForm +=
                integrate( _range=elements(mesh),
                           _expr= 0.5*idv(rho)*idv(this->velocityDiv())*trans(idv(betaU))*id(v),
                           _geomap=this->geomap() );
        }
    }

    //--------------------------------------------------------------------------------------------------//

    this->updateLinearPDEStabilisation(A,F,_BuildCstPart,A_extended,_BuildExtendedPart);

    //--------------------------------------------------------------------------------------------------//
    // neumann condition
    if (build_BoundaryNeumannTerm)
    {
        this->updateBCNeumannLinearPDE( F );
    }
    //pressure fix condition
    if (BuildCstPart && !this->markerPressureBC().empty() )
    {
        bilinearForm_PatternCoupled +=
            integrate( _range=markedfaces(mesh,this->markerPressureBC() ),
                       _expr= -trans(2*idv(this->densityViscosityModel()->fieldMu())*deft*N())*id(v),
                       _geomap=this->geomap() );
    }
    if (BuildNonCstPart)
    {
        this->updateBCPressureLinearPDE( F );
    }
    // others bc
    this->updateLinearPDEWeakBC(A,F,_BuildCstPart);

    //--------------------------------------------------------------------------------------------------//
    // strong Dirichlet bc
    if ( BuildNonCstPart && _doBCStrongDirichlet)
    {
        this->timerTool("Solve").start();

        if (this->hasMarkerDirichletBCelimination() )
            this->updateBCStrongDirichletLinearPDE(A,F);

#if defined( FEELPP_MODELS_HAS_MESHALE )
        if (this->isMoveDomain() && this->couplingFSIcondition()=="dirichlet-neumann")
        {
            bilinearForm_PatternCoupled +=
                on( _range=markedfaces(this->mesh(),this->markersNameMovingBoundary()),
                    _element=u, _rhs=F,
                    _expr=idv(this->meshVelocity2()) );
        }
#endif

        for ( auto const& inletbc : M_fluidInletDesc )
        {
            std::string const& marker = std::get<0>( inletbc );
            auto const& inletVel = std::get<0>( M_fluidInletVelocityInterpolated.find(marker)->second );

            bilinearForm_PatternCoupled +=
                on( _range=markedfaces(this->mesh(),marker),
                    _element=u, _rhs=F,
                    _expr=-idv(inletVel)*N() );
        }

        double timeElapsedDirichletBC = this->timerTool("Solve").stop();
        this->log("FluidMechanics","updateLinearPDE","assembly strong DirichletBC in "+(boost::format("%1% s") %timeElapsedDirichletBC).str() );
    }

    double timeElapsed = this->timerTool("Solve").stop();
    this->log("FluidMechanics","updateLinearPDE","finish in "+(boost::format("%1% s") %timeElapsed).str() );

} // updateLinearPDE

} // end namespace FeelModels
} // end namespace Feel


